#include <ev.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <poll.h>
#include <ctype.h>
#include <libgen.h>
#include <time.h>
#include <errno.h>
#include <stddef.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <net/if.h>

#include <unistd.h>
#include <fcntl.h>

#include <linux/can.h>
#include <linux/can/raw.h>
#include "lib.h"

#define SECOND (1000000LL)
#define MSECOND (1000LL)

static int can_sender_id = 0x123;
static int can_sender_type = CAN_EFF_FLAG;
static int can_sender_len = 8;
static char can_data[] = { 0, 1, 2, 3, 4, 5, 6, 7 };
static int can_delay_us = 0;
static int rx_process_time = 30;

static void time_start(struct timeval *tv)
{
	gettimeofday(tv, NULL);
}

static time_t time_delta(struct timeval *start, struct timeval *stop)
{
	time_t secs, usecs;

	secs = stop->tv_sec - start->tv_sec;
	usecs = stop->tv_usec - start->tv_usec;
	if (usecs < 0) {
		secs--;
		usecs += SECOND;
	}

	return secs * SECOND + usecs;
}

static time_t time_stop(struct timeval *begin)
{
	struct timeval end;
	gettimeofday(&end, NULL);
	return time_delta(begin, &end);
}

static int init_canbus(const char *device)
{

	struct sockaddr_can addr;
	struct ifreq ifr;
	int s;
	int loopback = 1;

	if ((s = socket(PF_CAN, SOCK_RAW, CAN_RAW)) < 0) {
		perror("socket");
		return -1;
	}

	addr.can_family = AF_CAN;
	strcpy(ifr.ifr_name, device);

	if (ioctl(s, SIOCGIFINDEX, &ifr) < 0) {
		perror("SIOCGIFINDEX");
		return -1;
	}

	addr.can_ifindex = ifr.ifr_ifindex;
	printf("%s: %d\n", device, ifr.ifr_ifindex);

	//setsockopt(s, SOL_CAN_RAW, CAN_RAW_FILTER, NULL, 0);
	setsockopt(s, SOL_CAN_RAW, CAN_RAW_LOOPBACK, &loopback, sizeof(loopback));

	if (bind(s, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
		perror("bind");
		return -1;
	}

	return s;
}

int is_can_tx(char *dev)
{
	char buf[32], data;
	int fd;
	int len;

	sprintf(buf, "/sys/class/net/%s/tx_empty", dev);

	fd = open(buf, O_RDONLY);
	if (fd < 0) {
		printf("open failed\n");
		return -1;
	}

	len = read(fd, &data, sizeof(data));
	if (len < 1) {
		printf("read failed\n");
		close(fd);
		return -1;
	}

	close(fd);

	if (data != '1')
		return -1;

	return 0;
}

void do_tx(int fd)
{
	struct can_frame frame;
	int nbytes, i;
	struct pollfd fds;
	uint64_t data = 0;

	memset(&frame, 0, sizeof(frame));
	memset(&fds, 0, sizeof(fds));

	fds.fd = fd;
	fds.events = POLLOUT;

	if (can_sender_type == CAN_EFF_FLAG) {
		frame.can_id = can_sender_id;
		frame.can_id &= CAN_EFF_MASK;
		frame.can_id |= CAN_EFF_FLAG;
	} else {
		frame.can_id = can_sender_id;
		frame.can_id &= CAN_SFF_MASK;
	}

	if (can_sender_len) {
		frame.can_dlc = can_sender_len;
		memcpy(frame.data, can_data, can_sender_len);
	} else {
		frame.can_id |= CAN_RTR_FLAG;
	}

	while (1) {
		while (1) {
			if (can_delay_us)
				usleep(can_delay_us);

			nbytes = write(fd, &frame, CAN_MTU);

			if (nbytes == CAN_MTU) {
				//data++;
				break;
			} else if (nbytes >= 0) {
				printf("%s pending: %d\n", __func__, nbytes);
				printf("%s pending: %d\n", __func__, nbytes);
				printf("%s pending: %d\n", __func__, nbytes);
				printf("%s pending: %d\n", __func__, nbytes);
				printf("%s pending: %d\n", __func__, nbytes);
			} else {
				if (errno == ENOBUFS || errno == EAGAIN)
					continue;

				printf("%s: error: %d %d\n", __func__, nbytes, errno);
				printf("%s: error: %d %d\n", __func__, nbytes, errno);
				printf("%s: error: %d %d\n", __func__, nbytes, errno);
				printf("%s: error: %d %d\n", __func__, nbytes, errno);
				exit(0);
			}
		}
	}
}

void do_rx(int fd)
{
	struct can_frame frame;
	int nbytes;
	uint64_t count = 0;
	struct timeval start, end;
	time_t elapse;
	float cur_rate, min_rate, max_rate;
	int wait_stable_cnt = 10;
	unsigned char buf[128];
	uint32_t can_id = 0;

	min_rate = max_rate = -1;

	while (1) {
		nbytes = read(fd, &frame, CAN_MTU);
		if (nbytes != CAN_MTU) {
			if (nbytes >= 0) {
				printf("%s pending: %d\n", __func__, nbytes);
				printf("%s pending: %d\n", __func__, nbytes);
				printf("%s pending: %d\n", __func__, nbytes);
				printf("%s pending: %d\n", __func__, nbytes);
				printf("%s pending: %d\n", __func__, nbytes);
			} else {

				if (EAGAIN == errno)
					continue;

				printf("%s: error: %d %d\n", __func__, nbytes, errno);
				printf("%s: error: %d %d\n", __func__, nbytes, errno);
				printf("%s: error: %d %d\n", __func__, nbytes, errno);
				printf("%s: error: %d %d\n", __func__, nbytes, errno);
				printf("%s: error: %d %d\n", __func__, nbytes, errno);

				exit(0);
			}
		}

		/* check id */
		if (!can_id) {
			if (frame.can_id & CAN_EFF_FLAG) {
				can_id = can_sender_id;
				can_id &= CAN_EFF_MASK;
				can_id |= CAN_EFF_FLAG;
			} else {
				can_id = can_sender_id;
				can_id &= CAN_SFF_MASK;
			}

			if (!frame.can_dlc)
				can_id |= CAN_RTR_FLAG;
		}

		if (frame.can_id != can_id) {
			printf("error packet id, %x, %x\n", frame.can_id, can_id);
			return;
		}

		/* check data */
		if (memcmp(frame.data, can_data, frame.can_dlc)) {
			//sprint_canframe(buf, &frame, 0, CAN_MAX_DLEN);
			//printf("error packet: %s", buf);
			printf("error packet data\n");
			return;
		}

		if (!count)
			time_start(&start);

		count++;

		if (count % 1000)
			continue;

		elapse = time_stop(&start);
		cur_rate = (count * SECOND * 1.0) / elapse;

		if (min_rate < 0)
			min_rate = cur_rate;
		else if (cur_rate < min_rate)
			min_rate = cur_rate;

		if (max_rate < 0)
			max_rate = cur_rate;
		else if (cur_rate > max_rate)
			max_rate = cur_rate;

		if (!wait_stable_cnt) {
#if 0
			printf("total: %zu, rx: %f/s, min/max: %f, %f/s\n", count, cur_rate,
			       min_rate, max_rate);
#else
			//printf("%f, %f, %f\n", cur_rate, min_rate, max_rate); 
			printf("%d, %d, %d\n", (int)cur_rate, (int)min_rate, (int)max_rate);
#endif
		} else {
			wait_stable_cnt--;
			printf("wait_stable_cnt: %d\n", wait_stable_cnt);
			time_start(&end);
		}

		elapse = time_stop(&end);
		if (rx_process_time && elapse >= rx_process_time * SECOND) {
			printf("After 30 sec\n");
			break;
		}

		count = 0;
		time_start(&start);
	}
}

int main(int argc, char *argv[])
{
	int fd;

	if (argc < 3) {
		printf("param failed\n");
		return 1;
	}

	fd = init_canbus(argv[1]);
	if (fd < 0) {
		printf("init fd failed\n");
		return 1;
	}

	if (argv[2][0] == '0') {
		if (argc != 6) {
			printf("param222 failed\n");
			return 1;
		}

		if (argv[3][0] == '0') {
			printf("tx sff\n");
			can_sender_type = 0;
		} else {
			printf("tx eff\n");
			can_sender_type = CAN_EFF_FLAG;
		}

		can_sender_len = argv[4][0] - '0';
		if (can_sender_len >= CAN_MAX_DLC)
			can_sender_len = CAN_MAX_DLC;

		printf("tx len: %d\n", can_sender_len);

		can_delay_us = strtol(argv[5], NULL, 10);
		printf("delay: %d\n", can_delay_us);

		printf("do tx\n");
		do_tx(fd);
	} else {
		if (argc >= 4) {
			rx_process_time = strtol(argv[3], NULL, 10);
			printf("rx_process_time: %d\n", rx_process_time);
		}
		
		printf("do rx\n");
		do_rx(fd);
	}

	return 0;
}
