#include <ev.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <poll.h>
#include <ctype.h>
#include <libgen.h>
#include <time.h>
#include <errno.h>
#include <stddef.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <net/if.h>

#include <unistd.h>
#include <fcntl.h>

#include <linux/can.h>
#include <linux/can/raw.h>
#include "lib.h"

#define SECOND (1000000LL)
#define MSECOND (1000LL)

static void time_start(struct timeval *tv)
{
	gettimeofday(tv, NULL);
}

static time_t time_delta(struct timeval *start, struct timeval *stop)
{
	time_t secs, usecs;

	secs = stop->tv_sec - start->tv_sec;
	usecs = stop->tv_usec - start->tv_usec;
	if (usecs < 0) {
		secs--;
		usecs += SECOND;
	}

	return secs * SECOND + usecs;
}

static time_t time_stop(struct timeval *begin)
{
	struct timeval end;
	gettimeofday(&end, NULL);
	return time_delta(begin, &end);
}

static int init_canbus(const char *device)
{

	struct sockaddr_can addr;
	struct ifreq ifr;
	int s;
	int loopback = 1;

	if ((s = socket(PF_CAN, SOCK_RAW, CAN_RAW)) < 0) {
		perror("socket");
		return -1;
	}

	addr.can_family = AF_CAN;
	strcpy(ifr.ifr_name, device);

	if (ioctl(s, SIOCGIFINDEX, &ifr) < 0) {
		perror("SIOCGIFINDEX");
		return -1;
	}

	addr.can_ifindex = ifr.ifr_ifindex;
	printf("%s: %d\n", device, ifr.ifr_ifindex);

	//setsockopt(s, SOL_CAN_RAW, CAN_RAW_FILTER, NULL, 0);
	setsockopt(s, SOL_CAN_RAW, CAN_RAW_LOOPBACK, &loopback, sizeof(loopback));

	if (bind(s, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
		perror("bind");
		return -1;
	}

	return s;
}

void dump_frame(struct can_frame *frame, struct can_frame *old_frame)
{
	unsigned char buf[128];
	sprint_canframe(buf, frame, 0, CAN_MAX_DLEN);
	printf("current RX: %s\n", buf);
	sprint_canframe(buf, old_frame, 0, CAN_MAX_DLEN);
	printf("old RX: %s\n", buf);
}

void do_tx(int fd)
{
	struct can_frame frame[9];
	int nbytes, i;
	struct pollfd fds;
	uint64_t data = 0;
	int max_udelay = 2200;
	int min_udelay = 1800;
	int cur_udelay = max_udelay;
	int dir = -1;
	struct timespec req, rem;
	int send_packet = sizeof(frame) / sizeof(struct can_frame);
	unsigned char tmp;

	memset(frame, 0, sizeof(frame));
	memset(&fds, 0, sizeof(fds));

	fds.fd = fd;
	fds.events = POLLOUT;

	for (i = 0; i < send_packet; ++i) {
		frame[i].can_id = (i << 16) | i;
		//frame[i].can_dlc = i % 9;
		frame[i].can_dlc = 0;
		memset(frame[i].data, i, sizeof(frame[i].data));
	}

	//return 0;
	while (1) {
		for (i = 0; i < send_packet; ++i) {
			nbytes = write(fd, &frame[i], CAN_MTU);

			if (nbytes == CAN_MTU) {
				continue;
			} else if (nbytes >= 0) {
				printf("%s pending: %d\n", __func__, nbytes);
			} else {
				if (errno == ENOBUFS || errno == EAGAIN) {
					printf("%s: error: %d %d continued\n", __func__, nbytes, errno);
					continue;
				}
				
				printf("%s: error: %d %d\n", __func__, nbytes, errno);
				exit(0);
			}
		}

		if (cur_udelay == max_udelay)
			dir = -1;

		if (cur_udelay == min_udelay)
			dir = 1;

		//usleep(10000);
		cur_udelay += dir;
		//printf("%d %d\n", cur_udelay, dir);
	}
}

void do_rx(int fd)
{
	struct can_frame frame, old_frame;
	int nbytes;
	uint64_t count = 0;
	struct timeval start, end;
	time_t elapse;
	float cur_rate, min_rate, max_rate;
	int wait_stable_cnt = 0;
	int is_first = 1, is_err;
	int err_cnt = 0;
	unsigned long id, data, old_id, old_data;

	min_rate = max_rate = -1;

	while (1) {
		is_err = 0;

		nbytes = read(fd, &frame, CAN_MTU);
		if (nbytes != CAN_MTU) {
			if (nbytes >= 0) {
				printf("%s pending: %d\n", __func__, nbytes);
				printf("%s pending: %d\n", __func__, nbytes);
				printf("%s pending: %d\n", __func__, nbytes);
				printf("%s pending: %d\n", __func__, nbytes);
				printf("%s pending: %d\n", __func__, nbytes);
			} else {

				if (EAGAIN == errno)
					continue;

				printf("%s: error: %d %d\n", __func__, nbytes, errno);
				printf("%s: error: %d %d\n", __func__, nbytes, errno);
				printf("%s: error: %d %d\n", __func__, nbytes, errno);
				printf("%s: error: %d %d\n", __func__, nbytes, errno);
				printf("%s: error: %d %d\n", __func__, nbytes, errno);

				exit(0);
			}
		}

		if (!count)
			time_start(&start);

		count++;

		// is first, copy and continue
		if (is_first) {
			is_first = 0;
			memcpy(&old_frame, &frame, sizeof(frame));
			continue;
		}
		// compare frame
		if (frame.can_id & CAN_ERR_FLAG) {
			printf("err flag err\n");
			dump_frame(&frame, &old_frame);
			is_err = 1;
		}

		if (frame.can_id & CAN_EFF_FLAG) {
			printf("eff flag err\n");
			dump_frame(&frame, &old_frame);
			is_err = 1;
		}

		if (memcmp(frame.data, old_frame.data, old_frame.can_dlc)) {
			printf("data not the same err\n");
			dump_frame(&frame, &old_frame);
			is_err = 1;
		}

		if (is_err) {
			err_cnt += 1;
			printf("err_cnt: %d\n", err_cnt);
		}
		// copy now to old
		memcpy(&old_frame, &frame, sizeof(frame));
		fflush(stdout);
#if 0
		if (count % 10000)
			continue;

		elapse = time_stop(&start);
		cur_rate = (count * SECOND) / elapse;

		if (min_rate < 0)
			min_rate = cur_rate;
		else if (cur_rate < min_rate)
			min_rate = cur_rate;

		if (max_rate < 0)
			max_rate = cur_rate;
		else if (cur_rate > max_rate)
			max_rate = cur_rate;

		if (wait_stable_cnt) {
			wait_stable_cnt--;
			printf("wait_stable_cnt: %d\n", wait_stable_cnt);
			time_start(&end);
			continue;
		}
		//printf("%d, %d, %d\n", (int)cur_rate, (int)min_rate, (int)max_rate);
		/*
		   elapse = time_stop(&end);
		   if (rx_process_time && elapse >= rx_process_time * SECOND) {
		   printf("After 30 sec\n");
		   break;
		   }
		 */
		count = 0;
		time_start(&start);
#endif
	}
}

int main(int argc, char *argv[])
{
	int fd;

	fd = init_canbus(argv[1]);
	if (fd < 0) {
		printf("init fd failed\n");
		return 1;
	}

	if (argc < 3) {
		printf("parameter error\n");
		return 0;
	}

	if (argv[2][0] == '0') {
		printf("do rx\n");
		fflush(stdout);
		do_rx(fd);
	} else {
		printf("do tx\n");
		fflush(stdout);
		do_tx(fd);
	}

	return 0;
}
